#include <windows.h>
#include "icinga-version.h"

LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US

100 ICON "icinga.ico"

VS_VERSION_INFO VERSIONINFO
FILEVERSION 1,0,0,0
PRODUCTVERSION 1,0,0,0
FILEOS VOS__WINDOWS32
FILETYPE VFT_APP
FILESUBTYPE VFT2_UNKNOWN
BEGIN
	BLOCK "StringFileInfo"
	BEGIN
		BLOCK "040904E4"
		BEGIN
			VALUE "CompanyName", "Icinga GmbH"
			VALUE "FileDescription", "Icinga 2"
			VALUE "FileVersion", ICINGA2_VERSION
			VALUE "InternalName", "icinga2.exe"
			VALUE "LegalCopyright", "© Icinga GmbH"
			VALUE "OriginalFilename", "icinga2.exe"
			VALUE "ProductName", "Icinga 2"
			VALUE "ProductVersion", ICINGA2_VERSION
		END
	END

	BLOCK "VarFileInfo"
	BEGIN
		VALUE "Translation", 0x409, 0x04E4
	END
END
