/* Icinga 2 | (c) 2012 Icinga GmbH | GPLv2+ */

#ifndef I2DB_IDO_H
#define I2DB_IDO_H

/**
 * @defgroup db_ido IDO library
 *
 * The Icinga library implements database-agnostic IDO functionality.
 */

#include "base/i2-base.hpp"

#endif /* I2DB_IDO_H */
